@@ -2,11 +2,18 @@ import json
 import boto3
 import logging
 import pandas as pd
+import os
+os.environ['MPLCONFIGDIR'] = "/tmp"
+import matplotlib.pyplot as plt
+# import seaborn as sns
 import numpy as np
+import math
 from io import BytesIO
 from datetime import datetime
 import statistics
 from botocore.exceptions import ClientError
+import numbers
+from matplotlib.colors import LogNorm
 
 
 def days_diff(d1, d2):
@@ -29,7 +36,7 @@ def accuracyMetric(target, reported):
     return abs(target - reported)
 
 
-def accuracy_metrics(vals, expected_dt, truth_val):
+def accuracy_metrics(vals, truth_val, expected_dt):
     """
 
     :param vals:
@@ -67,7 +74,7 @@ def accuracy_metrics(vals, expected_dt, truth_val):
     }
 
 
-def completeness_metric(series,  truth_df, expected_df, *args, **kwargs):
+def completeness_metric(series, truth_df, expected_df, *args, **kwargs):
     """
     Returns percent of an update that includes previous values
 
@@ -78,11 +85,15 @@ def completeness_metric(series,  truth_df, expected_df, *args, **kwargs):
     """
     # check that there is an update for all the expected less than or equal to this
     expected_vals = expected_df[expected_df.iloc[:, 0] < series.name].index
-    num_missing = len([each for each in expected_vals if each not in series.index])
-    return {"Completeness": 1 - (num_missing / len(expected_vals))}
+    not_na = series[series.notna()]
+    num_missing = len([each for each in expected_vals if each not in not_na.index])
+    if len(expected_vals) > 0:
+        return {"Completeness": 1 - (num_missing / len(expected_vals))}
+    else:
+        return {"Completeness": None}
 
 
-def time_till_metrics(vals, expected_dt, truth_val):
+def time_till_metrics(vals, truth_val, expected_dt):
     """
 
     :param vals: pandas Series object
@@ -98,28 +109,37 @@ def time_till_metrics(vals, expected_dt, truth_val):
     first_reported = vals_not_na.index[0]
 
     # ------Time Till Accurate------- #
+    # Else the final value was correct, and we need to find the last date for which it changed
+
+    # changes between updates; [NaN, val1-val0, ...]
+    diffs = [each for each in vals_not_na.diff()]
+
+    # Find the dates at which the value changed
+    change_dates = [ind for ind, dff in zip(vals_not_na.index, diffs) if dff != 0]
+    time_till_accurate = days_diff(expected_dt, change_dates[-1])
+
     # If final value was not correct
     if truth_val != vals_not_na[-1]:
-        time_till_accurate = np.Inf  #  should this be something like more days than have elapsed?
-
-    else:
-        # Else the final value was correct, and we need to find the last date for which it changed
-
-        # changes between updates; [NaN, val1-val0, ...]
-        diffs = [each for each in vals_not_na.diff()]
-
-        # Find the dates at which the value changed
-        change_dates = [ind for ind, dff in zip(vals_not_na.index, diffs) if dff != 0]
-        time_till_accurate = days_diff(expected_dt, change_dates[-1])
+        time_till_accurate = 2 * days_diff(expected_dt,
+                                           vals.index[-1])  # should this be something like more days than have elapsed?
     return {
         "TimeTillReported": days_diff(expected_dt, first_reported),
         "TimeTillAccurate": time_till_accurate
     }
 
 
-def validity_nondecreasing_smoothness(series,  truth_df, expected_df, *args, **kwargs):
+def validity_nondecreasing_step(series, truth_df, expected_df, *args, **kwargs):
+    truth_max = truth_df[expected_df.index < series.name].iloc[:, 0].max()
+    num_expected_updates = len(expected_df[expected_df.index < series.name].iloc[:, 0].unique())
+    coef = -1 / (min(num_expected_updates, truth_max) / 5)
+    num_increasing_updates = len(series[series.diff() > 0])
+
+    return -1 + 2 / (1 + math.exp(coef * num_increasing_updates))
+
+
+def validity_nondecreasing_smoothness(series, truth_df, expected_df, *args, **kwargs):
     """
-    Assumes series is of increasing dates by nondecreasing observations
+    Assumes series is of nondecreasing observations by increasing dates
 
     :param series:
     :param truth_df:
@@ -128,37 +148,70 @@ def validity_nondecreasing_smoothness(series,  truth_df, expected_df, *args, **k
     :param kwargs:
     :return:
     """
-    orig_diffs = series.diff()
-    rises = [abs(diff) for diff in orig_diffs if diff > 0]
-    inds = [ind for ind, diff in enumerate(orig_diffs) if diff > 0]
+    orig_diffs = series.diff()  # nan, 2
+    rises = [abs(diff) for diff in orig_diffs if diff > 0]  # 2
+    inds = [ind for ind, diff in enumerate(orig_diffs) if diff > 0]  # 2020-03-28
 
     if len(inds) == 0:
         return 0
 
-    runs = [inds[0]]
-    runs = runs + [abs(ind2 - ind1) for ind1, ind2 in zip(inds[:-1], inds[1:])]
+    runs = [inds[0]]  # 2020-03-28
+    if len(inds) > 1:
+        runs = runs + [abs(ind2 - ind1) for ind1, ind2 in zip(inds[:-1], inds[1:])]
 
     # last observation updated (max dt not na)
     obsvs = series[series.notna()]
-    max_date = obsvs.index[-1]
-    max_val = obsvs.iloc[-1][0]
-    num_expected_dates = expected_df[expected_df.iloc[:, 0] < max_date].shape[0]
-    total_area = max_val * num_expected_dates
-    covered_area = sum([rise * run for rise, run in zip(rises, runs)])
-    return 1 - (covered_area / total_area)
+    # max_date = obsvs.index[-1]
+    max_val = obsvs.iloc[-1]  # 6
+    num_expected_dates = expected_df[expected_df.iloc[:, 0] < series.name].shape[0]  # 2
+    total_area = max_val * num_expected_dates  # 6x2
+    covered_area = sum([rise * run for rise, run in zip(rises, runs)])  # sum([]
+    if total_area > 0:
+        return 1 - (covered_area / total_area)
+    else:
+        return None
 
 
 def validity_nondecreasing(series, *args, **kwargs):
-    """Checks if values are nondecreasing.
-    Return number of times a decrease occurs"""
+    """Percent of sequential reported dates that decrease"""
     num_decreasing_updates = len([each for each in series[series.notna()].diff() if each < 0])
-    num_updates = len(series[series.notna()])
-    if num_updates > 1:
+    num_updates = len([each for each in series[series.notna()].diff() if each != 0])
+    if num_updates > 2:
         return 1 - (num_decreasing_updates / (num_updates - 1))
     else:
         return 1
 
 
+def consistency_nondecreasing_percent_restatement(series, truth_df, expected_df, *args, **kwargs):
+    # |sum of increases| / |sum of absolute changes|
+    increases = sum([diff for diff in series.diff().dropna() if diff > 0])
+    changes = sum([abs(diff) for diff in series.diff().dropna()])
+    if changes > 0:
+        return increases / changes
+    else:
+        return None
+
+
+def consistency_between_updates(data_df, truth_df, expected_df):
+    updates = data_df.columns
+    dct = {}
+    last_non_null = None
+
+    for prior, post in zip(updates[:-1], updates[1:]):
+
+        if len(data_df[prior].notna()) > 0:
+            last_non_null = data_df[prior]
+        post_nonnulls = data_df[post].notna()
+        diff = data_df[post][post_nonnulls] - last_non_null[post_nonnulls]
+
+        if len(diff) > 0:
+            dct[post] = {"ConsistencyPercentUpdated": len(diff[diff > 0]) / len(diff)}
+        else:
+            dct[post] = None
+
+    return pd.DataFrame.from_dict(dct)
+
+
 # ----------------------------------------------#
 # -------- Data Quality Dimensions--------------#
 # ----------------------------------------------#
@@ -183,7 +236,7 @@ def apply_to_observations(func):
                 truth_val = truth_df.loc[obsv][0]
                 expected_dt = expected_df.loc[obsv][0]
 
-                dct[obsv] = func(vals, expected_dt, truth_val, *args, **kwargs)
+                dct[obsv] = func(vals, truth_val, expected_dt, *args, **kwargs)
 
         return pd.DataFrame.from_dict(dct, orient='index')
 
@@ -246,7 +299,9 @@ def validity(data_df, truth_df, expected_df):
             "ValidityNondecreasingSmoothness":
                 validity_nondecreasing_smoothness(series, truth_df, expected_df),
             "ValidityNondecreasingUpdates":
-                validity_nondecreasing(series, truth_df, expected_df)
+                validity_nondecreasing(series, truth_df, expected_df),
+            "ValidityNondecreasingStep":
+                validity_nondecreasing_step(series, truth_df, expected_df),
         }
 
     func = apply_to_updates(validity_metrics)
@@ -254,33 +309,50 @@ def validity(data_df, truth_df, expected_df):
 
 
 def consistency(data_df, truth_df, expected_df):
-    updates = data_df.columns
-    dct = {}
-    last_non_null = None
-    for prior, post in zip(updates[:-1], updates[1:]):
+    def consistency_metrics(series, truth_df, expected_df):
+        return {
+            "ConsistencyNondecreasingPercentRestatement":
+                consistency_nondecreasing_percent_restatement(series, truth_df, expected_df)
+        }
 
-        if len(data_df[prior].notna()) > 0:
-            last_non_null = data_df[prior]
-        post_nonnulls = data_df[post].notna()
-        diff = data_df[post][post_nonnulls] - last_non_null[post_nonnulls]
-
-        if len(diff) > 0:
-            dct[post] = {"ConsistencyPercentUpdated": len(diff[diff > 0]) / len(diff)}
-        else:
-            dct[post] = None
-
-    return pd.DataFrame.from_dict(dct)
+    func = apply_to_updates(consistency_metrics)
+    df1 = func(data_df, truth_df, expected_df)
+    print(type(df1))
+    print(df1.shape)
+    df2 = consistency_between_updates(data_df, truth_df, expected_df)
+    print(type(df2))
+    print(df2.shape)
+    return pd.concat([df1, df2])
 
 
 def completeness(data_df, truth_df, expected_df):
-    func = apply_to_observations(completeness_metric)
+    func = apply_to_updates(completeness_metric)
     return func(data_df, truth_df, expected_df)
 
 
+def basic_stats(data_df, truth_df, expected_df):
+    desc = truth_df.describe().T
+    desc.columns = [f'truth_{each}' for each in desc.columns]
+    return desc
+
+
 # ----------------------------------------------#
 # -------------- AWS Management-----------------#
 # ----------------------------------------------#
 
+def save_plt_to_s3(plt, bucket, key, format='png'):
+    img_data = BytesIO()
+    plt.savefig(img_data, format=format)
+    img_data.seek(0)
+
+    s3 = boto3.resource('s3')
+    bucket = s3.Bucket(bucket)
+    print(f'Attempting to put image {key} in {bucket}')
+    resp = bucket.put_object(Body=img_data, ContentType='image/png', Key=key)
+    print(f'Successfully put image {key} in {bucket}')
+    return resp
+
+
 def df_from_s3_csv(bucket, key):
     """
 
@@ -313,7 +385,6 @@ def df_to_s3_csv(bucket, key, df):
 
     print("Writing BytesIO")
     with BytesIO() as csv_buffer:
-
         df.to_csv(csv_buffer)
         print(f'Attempting to put object to bucket {bucket} at key {key}')
         response = s3_client.put_object(Bucket=bucket, Key=key, Body=csv_buffer.getvalue())
@@ -342,7 +413,16 @@ def send_sqs_message(to_sqs_url, msg_body, region_name):
         logging.error(e)
         return None
 
+def make_key(file, path=''):
+    if path == "":
+        return file
+    else:
+        return "/".join([path, file])
 
+
+# ----------------------------------------------#
+# -------------- Nodes -------------------------#
+# ----------------------------------------------#
 def metric_node(
         event,
         context,
@@ -370,6 +450,7 @@ def metric_node(
     :return:
     """
     funcs_mapping = {
+        "stats": basic_stats,
         "accuracy": accuracy,
         "timeliness": timeliness,
         "validity": validity,
@@ -396,21 +477,15 @@ def metric_node(
     if 'expected_file' in body:
         expected_file = body['expected_file']
 
-    def make_key(file):
-        if path == "":
-            return file
-        else:
-            return "/".join([path, file])
-
     completed_stations = [each for each in body['completed_stations']]
 
-    data_df = df_from_s3_csv(from_bucket, make_key(data_file))
-    truth_df = df_from_s3_csv(from_bucket, make_key(truth_file))
-    expected_df = df_from_s3_csv(from_bucket, make_key(expected_file))
+    data_df = df_from_s3_csv(from_bucket, make_key(data_file, path))
+    truth_df = df_from_s3_csv(from_bucket, make_key(truth_file, path))
+    expected_df = df_from_s3_csv(from_bucket, make_key(expected_file, path))
 
     finished_df = funcs_mapping[func](data_df, truth_df, expected_df)
 
-    status = df_to_s3_csv(bucket=to_bucket, key=make_key(to_key), df=finished_df)
+    status = df_to_s3_csv(bucket=to_bucket, key=make_key(to_key, path), df=finished_df)
 
     if status == 200:
         logging.info(f'Successful S3 put_object response')
@@ -459,9 +534,6 @@ def summary_node(
 
 
     """
-    logging.getLogger().setLevel(logging.INFO)
-    # logging.basicConfig(level=logging.DEBUG,
-    #                     format='%(levelname)s: %(asctime)s: %(message)s')
     dims_files = [
         "timeliness.csv",
         "accuracy.csv",
@@ -484,21 +556,13 @@ def summary_node(
     if 'expected_file' in body:
         expected_file = body['expected_file']
 
-    # from_bucket = body['bucket']  # unless completed edits the event, this is prepared-data
-
-    def make_key(file):
-        if path == "":
-            return file
-        else:
-            return "/".join([path, file])
-
     completed_stations = [each for each in body['completed_stations']]
 
     dfs = []
-    transpose = ["validity.csv", "consistency.csv"]
+    transpose = ["validity.csv", "consistency.csv", "completeness.csv"]
 
     for from_key in dims_files:
-        df = df_from_s3_csv(from_bucket, make_key(from_key))
+        df = df_from_s3_csv(from_bucket, make_key(from_key, path))
 
         if from_key in transpose:
             df = df.T
@@ -516,13 +580,18 @@ def summary_node(
 
     finished_df = pd.concat(dfs)
 
-    status = df_to_s3_csv(bucket=to_bucket, key=make_key(to_key), df=finished_df)
+    status = df_to_s3_csv(bucket=to_bucket, key=make_key(to_key, path), df=finished_df)
 
-    if status == 200:
-        logging.info(f'Successful S3 put_object response')
+    flattened = finished_df.unstack().to_frame().sort_index(level=1).T
+    flattened.columns = flattened.columns.map('_'.join)
+    flattened.index = ['0']
 
-    else:
-        logging.info(f'Unsuccessful S3 put_object response - {status}')
+    stats = df_from_s3_csv(from_bucket, make_key("stats.csv", path))
+    stats.index = ['0']
+    flattened = pd.concat([stats, flattened], axis=1)
+    # concat flattened to stats
+
+    status = df_to_s3_csv(bucket=to_bucket, key=make_key(f'flattened_{to_key}', path), df=flattened)
 
     completed_stations.append(context.function_name)
 
@@ -545,3 +614,136 @@ def summary_node(
         'statusCode': status,
         'body': message
     }
+
+
+def plot_node(
+        event,
+        context,
+        to_sqs_url,
+        region_name,
+        from_bucket,
+        to_bucket,
+        to_key
+):
+    """
+
+    :param event:
+    :param context:
+    :param to_sqs_url:
+    :param region_name:
+    :param from_bucket:
+    :param to_bucket:
+    :param to_key:
+    :return:
+    """
+    body = event["Records"][0]["body"]
+    if isinstance(body, str):
+        body = json.loads(body)
+
+    from_bucket = body['bucket']
+    path, data_file, truth_file, expected_file = "", "data.csv", "truth.csv", "expected_update.csv"
+    summary_file = 'summary.csv'
+    prepared_bucket = 'prepared-data'
+    evaluated_bucket = 'evaluated-data'
+    if 'path' in body:
+        path = body['path']
+    if 'data_file' in body:
+        data_file = body['data_file']
+    if 'truth_file' in body:
+        truth_file = body['truth_file']
+    if 'expected_file' in body:
+        expected_file = body['expected_file']
+    if 'summary_file' in body:
+        summary_file = body['summary_file']
+
+    completed_stations = [each for each in body['completed_stations']]
+
+    # plot truth data
+    print("Creating Line Plot of Truth Values")
+    truth_df = df_from_s3_csv(prepared_bucket, make_key(truth_file, path))
+    image_name = 'truth_plot.png'
+    fancy_path = path.replace("/", ": ").replace("_", " ")
+    truth_df.iloc[:, -1].plot(kind="line")
+    plt.title(fancy_path, fontsize=15)
+    save_plt_to_s3(plt, to_bucket, make_key(image_name, path), format='png')
+    plt.clf()
+    print('Cleared Image')
+
+    # plot heatmap
+    print("Creating Heatmap of Data")
+    image_name = 'data_heatmap.png'
+    data_df = df_from_s3_csv(from_bucket, make_key(data_file, path))
+    plt.pcolormesh(data_df, norm=LogNorm())
+    len_ind = len(data_df.index)
+    len_cols = len(data_df.columns)
+    ynum = max(len_ind // 10, 1)
+    xnum = max(len_cols // 10, 1)
+    yticks = [each for each in range(len_ind) if each % ynum == 0]
+    xticks = [each for each in range(len_cols) if each % xnum == 0]
+    plt.yticks(ticks=[y + .5 for y in yticks], labels=data_df.index[yticks])
+    plt.xticks(ticks=[x + .5 for x in xticks], labels=data_df.columns[xticks])
+    plt.colorbar()
+    # sns.set(rc={'figure.figsize': (10, 10)})
+    # if data_df.max().max() > 50:
+    #     temp_data = data_df.fillna(.00001).replace(0, .1)
+    #     ax = sns.heatmap(temp_data, norm=LogNorm(clip=True))
+    # else:
+    #     temp_data = data_df.fillna(-1)
+    #     ax = sns.heatmap(temp_data)
+    #
+    plt.title(f'{fancy_path}', fontsize=15)
+    plt.xlabel('Update', fontsize=15)
+    plt.ylabel('Reporting Date (Subject)', fontsize=15)
+    save_plt_to_s3(plt, to_bucket, make_key(image_name, path), format='png')
+    plt.clf()
+    print('Cleared Image')
+
+    # create an image of the table
+    print("Creating Summary Table")
+    image_name = 'summary_table.png'
+    summary_df = df_from_s3_csv(evaluated_bucket, make_key(summary_file, path))
+
+    def shorten(x):
+        if isinstance(x, numbers.Number):
+            return round(x, 3)
+        else:
+            return x
+
+    column_headers = summary_df.columns
+    row_headers = summary_df.index
+
+    cell_text = []
+    for __, row in summary_df.iterrows():
+        cell_text.append([f'{shorten(x)}' for x in row])
+
+    plt.figure(linewidth=2,
+               figsize=(20, 9)
+               )
+    plt.tight_layout()
+    # Add a table at the bottom of the axes
+    the_table = plt.table(cellText=cell_text,
+                          rowLabels=row_headers,
+                          rowLoc='right',
+                          colLabels=column_headers,
+                          loc='center'
+                          )
+    the_table.auto_set_font_size(False)
+    the_table.set_fontsize(12)
+    the_table.scale(2, 2)
+    the_table.auto_set_column_width(col=list(range(len(summary_df.columns))))
+
+    ax = plt.gca()
+    ax.get_xaxis().set_visible(False)
+    ax.get_yaxis().set_visible(False)
+    # Hide axes border
+    plt.box(on=None)
+    # Add title
+    plt.suptitle(fancy_path, fontsize=20)
+    plt.draw()
+    save_plt_to_s3(plt, to_bucket, make_key(image_name, path), format='png')
+    plt.clf()
+
+    return {
+        'statusCode': 'temp_status',
+        'body': 'done'
+    }
\ No newline at end of file
